import { __rest, __spreadArray } from "tslib";
import React from 'react';
import { autocompletion, closeBrackets, closeBracketsKeymap, completionKeymap } from '@codemirror/autocomplete';
import { defaultKeymap, history, historyKeymap } from '@codemirror/commands';
import { cpp } from '@codemirror/lang-cpp';
import { markdown } from '@codemirror/lang-markdown';
import { javascript } from '@codemirror/lang-javascript';
import { html } from '@codemirror/lang-html';
import { css } from '@codemirror/lang-css';
import { xml } from '@codemirror/lang-xml';
import { java } from '@codemirror/lang-java';
import { json } from '@codemirror/lang-json';
import { less } from '@codemirror/lang-less';
import { php } from '@codemirror/lang-php';
import { python } from '@codemirror/lang-python';
import { sql } from '@codemirror/lang-sql';
import { rust } from '@codemirror/lang-rust';
import { wast } from '@codemirror/lang-wast';
import { sass } from '@codemirror/lang-sass';
import { bracketMatching, defaultHighlightStyle, foldGutter, foldKeymap, indentOnInput, syntaxHighlighting } from '@codemirror/language';
import { lintKeymap } from '@codemirror/lint';
import { highlightSelectionMatches, searchKeymap } from '@codemirror/search';
import { EditorState, Compartment } from '@codemirror/state';
import { crosshairCursor, drawSelection, dropCursor, EditorView, highlightActiveLine, highlightActiveLineGutter, highlightSpecialChars, keymap, lineNumbers, rectangularSelection } from '@codemirror/view';
import className from './style.module.css';
var languageModes = {
    h: cpp(),
    c: cpp(),
    clj: cpp(),
    coffee: cpp(),
    cpp: cpp(),
    cs: cpp(),
    css: css(),
    groovy: cpp(),
    haxe: cpp(),
    htm: html(),
    html: html(),
    java: java(),
    js: javascript(),
    ts: javascript(),
    jsm: javascript(),
    json: json(),
    less: less(),
    lua: cpp(),
    markdown: markdown(),
    md: markdown(),
    mdown: markdown(),
    mdwn: markdown(),
    mkd: markdown(),
    ml: cpp(),
    mli: cpp(),
    pl: cpp(),
    php: php(),
    py: python(),
    rb: cpp(),
    rs: rust(),
    sh: cpp(),
    sql: sql(),
    svg: xml(),
    sass: sass(),
    xml: xml(),
    wast: wast()
    //filetype.txt = 'txt';
};
function detectMode(ext) {
    var mode = languageModes[ext];
    if (mode) {
        return mode;
    }
    return undefined;
    // return autoFold(autoMode());
}
// https://codemirror.net/examples/config/#dynamic-configuration
var lineNumbersCompartment = new Compartment();
export function makeLineNumbers(showLineNumbers) {
    var v = showLineNumbers ? lineNumbers() : [];
    return lineNumbersCompartment.of(v);
}
var basicSetup = [
    highlightActiveLineGutter(),
    highlightSpecialChars(),
    history(),
    foldGutter(),
    drawSelection(),
    dropCursor(),
    EditorState.allowMultipleSelections.of(true),
    indentOnInput(),
    syntaxHighlighting(defaultHighlightStyle, { fallback: true }),
    bracketMatching(),
    closeBrackets(),
    autocompletion(),
    rectangularSelection(),
    crosshairCursor(),
    highlightActiveLine(),
    highlightSelectionMatches(),
    keymap.of(__spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray(__spreadArray([], closeBracketsKeymap, true), defaultKeymap, true), searchKeymap, true), historyKeymap, true), foldKeymap, true), completionKeymap, true), lintKeymap, true))
];
var createCodeMirrorState = function (_a) {
    var onChange = _a.onChange, lock = _a.lock, dark = _a.dark, editable = _a.editable, value = _a.value, fileExtension = _a.fileExtension;
    var mode = detectMode(fileExtension);
    return EditorState.create({
        doc: value,
        extensions: __spreadArray(__spreadArray([
            basicSetup,
            //https://github.com/kjk/onlinetool.io/blob/752dc581d6e7f38ddd3b28be1feed65e9de860b1/web/CodeMirrorConfig.js#L12
            makeLineNumbers(true)
        ], (mode !== undefined ? [mode] : []), true), [
            EditorView.updateListener.of(function (v) {
                if (v.focusChanged) {
                    lock.current = v.view.hasFocus;
                }
                if (v.docChanged) {
                    var getString = function () { return v.state.doc.toString(); };
                    onChange(getString);
                }
            }),
            EditorView.theme({
                '&.cm-focused': {
                    outline: 'none'
                }
            }, { dark: dark }),
            // https://github.com/codemirror/codemirror.next/issues/173
            EditorView.editable.of(editable)
        ], false)
    });
};
var createCodeMirrorView = function (_a) {
    var root = _a.root, options = __rest(_a, ["root"]);
    return new EditorView({
        state: createCodeMirrorState(options),
        parent: root
    });
};
export var CodeMirror = React.forwardRef(function (_a, ref) {
    var value = _a.value, onChange = _a.onChange, lock = _a.lock, dark = _a.dark, editable = _a.editable, fileExtension = _a.fileExtension;
    var divRef = React.useRef(null);
    var showLNumbers = React.useRef(true);
    var editorRef = React.useRef();
    var _b = React.useState(false), focus = _b[0], setFocus = _b[1];
    React.useEffect(function () {
        if (!divRef.current)
            return;
        var editor = createCodeMirrorView({
            root: divRef.current,
            onChange: onChange,
            lock: lock,
            dark: dark,
            editable: editable,
            value: value,
            fileExtension: fileExtension
        });
        editorRef.current = editor;
        /*const timer = setInterval(() => {
          editor.focus();
          if(editor.hasFocus) clearInterval(timer);
        }, 500);*/
        return function () {
            // clearInterval(timer);
            editor.destroy();
        };
    }, [onChange, value, lock, dark]);
    React.useImperativeHandle(ref, function () { return ({
        update: function (value) {
            var current = editorRef.current;
            if (!current)
                return;
            current.setState(createCodeMirrorState({
                onChange: onChange,
                lock: lock,
                dark: dark,
                editable: editable,
                value: value
            }));
        },
        toggleLineNumbers: function () {
            var current = editorRef.current;
            if (!current)
                return;
            showLNumbers.current = !showLNumbers.current;
            var v = showLNumbers.current ? lineNumbers() : [];
            current.dispatch({
                effects: lineNumbersCompartment.reconfigure(v)
            });
        }
    }); });
    return (React.createElement("div", { onFocus: function () { return setFocus(true); }, onBlur: function () { return setFocus(false); }, className: className['code'] + (focus ? ' ' + className['focus'] : '') },
        React.createElement("div", { ref: divRef })));
});
//# sourceMappingURL=CodeMirror.js.map